@@ -614,6 +614,79 @@ static int dump_to_cache(FILE *f, char *domain, char *path, struct exportent *ex
 	return qword_eol(f);
 }
 
+static int is_subdirectory(char *subpath, char *path)
+{
+	int l = strlen(path);
+
+	return strcmp(subpath, path) == 0
+		|| (strncmp(subpath, path, l) == 0 && path[l] == '/');
+}
+
+static int path_matches(nfs_export *exp, char *path)
+{
+	if (exp->m_export.e_flags & NFSEXP_CROSSMOUNT)
+		return is_subdirectory(path, exp->m_export.e_path);
+	return strcmp(path, exp->m_export.e_path) == 0;
+}
+
+static int client_matches(nfs_export *exp, char *dom, struct hostent *he)
+{
+	if (use_ipaddr)
+		return client_check(exp->m_client, he);
+	return client_member(dom, exp->m_client->m_hostname);
+}
+
+static int export_matches(nfs_export *exp, char *dom, char *path, struct hostent *he)
+{
+	return path_matches(exp, path) && client_matches(exp, dom, he);
+}
+
+static nfs_export *lookup_export(char *dom, char *path, struct hostent *he)
+{
+	nfs_export *exp;
+	nfs_export *found = NULL;
+	int found_type = 0;
+	int i;
+
+	for (i=0 ; i < MCL_MAXTYPES; i++) {
+		for (exp = exportlist[i].p_head; exp; exp = exp->m_next) {
+			if (!export_matches(exp, dom, path, he))
+				continue;
+			if (!found) {
+				found = exp;
+				found_type = i;
+				continue;
+			}
+
+			/* Always prefer non-V4ROOT mounts */
+			if (found->m_export.e_flags & NFSEXP_V4ROOT)
+				continue;
+
+			/* If one is a CROSSMOUNT, then prefer the longest path */
+			if (((found->m_export.e_flags & NFSEXP_CROSSMOUNT) ||
+			     (exp->m_export.e_flags & NFSEXP_CROSSMOUNT)) &&
+			    strlen(found->m_export.e_path) !=
+			    strlen(exp->m_export.e_path)) {
+
+				if (strlen(exp->m_export.e_path) >
+				    strlen(found->m_export.e_path)) {
+					found = exp;
+					found_type = i;
+				}
+				continue;
+
+			} else if (found_type == i && found->m_warned == 0) {
+				xlog(L_WARNING, "%s exported to both %s and %s, "
+				     "arbitrarily choosing options from first",
+				     path, found->m_client->m_hostname, exp->m_client->m_hostname,
+				     dom);
+				found->m_warned = 1;
+			}
+		}
+	}
+	return found;
+}
+
 void nfsd_export(FILE *f)
 {
 	/* requests are:
@@ -623,10 +696,8 @@ void nfsd_export(FILE *f)
 	 */
 
 	char *cp;
-	int i;
 	char *dom, *path;
-	nfs_export *exp, *found = NULL;
-	int found_type = 0;
+	nfs_export *found = NULL;
 	struct in_addr addr;
 	struct hostent *he = NULL;
 
@@ -650,60 +721,14 @@ void nfsd_export(FILE *f)
 
 	auth_reload();
 
-	/* now find flags for this export point in this domain */
-	for (i=0 ; i < MCL_MAXTYPES; i++) {
-		for (exp = exportlist[i].p_head; exp; exp = exp->m_next) {
-			if (!use_ipaddr && !client_member(dom, exp->m_client->m_hostname))
-				continue;
-			if (exp->m_export.e_flags & NFSEXP_CROSSMOUNT) {
-				/* if path is a mountpoint below e_path, then OK */
-				int l = strlen(exp->m_export.e_path);
-				if (strcmp(path, exp->m_export.e_path) == 0 ||
-				    (strncmp(path, exp->m_export.e_path, l) == 0 &&
-				     path[l] == '/' &&
-				     is_mountpoint(path)))
-					/* ok */;
-				else
-					continue;
-			} else if (strcmp(path, exp->m_export.e_path) != 0)
-				continue;
-			if (use_ipaddr) {
-				if (he == NULL) {
-					if (!inet_aton(dom, &addr))
-						goto out;
-					he = client_resolve(addr);
-				}
-				if (!client_check(exp->m_client, he))
-					continue;
-			}
-			if (!found) {
-				found = exp;
-				found_type = i;
-				continue;
-			}
-			/* If one is a CROSSMOUNT, then prefer the longest path */
-			if (((found->m_export.e_flags & NFSEXP_CROSSMOUNT) ||
-			     (exp->m_export.e_flags & NFSEXP_CROSSMOUNT)) &&
-			    strlen(found->m_export.e_path) !=
-			    strlen(exp->m_export.e_path)) {
-
-				if (strlen(exp->m_export.e_path) >
-				    strlen(found->m_export.e_path)) {
-					found = exp;
-					found_type = i;
-				}
-				continue;
-
-			} else if (found_type == i && found->m_warned == 0) {
-				xlog(L_WARNING, "%s exported to both %s and %s, "
-				     "arbitrarily choosing options from first",
-				     path, found->m_client->m_hostname, exp->m_client->m_hostname,
-				     dom);
-				found->m_warned = 1;
-			}
-		}
+	if (use_ipaddr) {
+		if (!inet_aton(dom, &addr))
+			goto out;
+		he = client_resolve(addr);
 	}
 
+	found = lookup_export(dom, path, he);
+
 	if (found) {
 		if (dump_to_cache(f, dom, path, &found->m_export) < 0) {
 			xlog(L_WARNING,
